'''Tests for canonizers'''
from collections import OrderedDict
from functools import partial

import pytest
import torch
from torch.nn import Sequential
from helpers import assert_identity_hook

from zennit.canonizers import Canonizer, CompositeCanonizer
from zennit.canonizers import SequentialMergeBatchNorm, NamedMergeBatchNorm, AttributeCanonizer
from zennit.core import RemovableHandleList
from zennit.types import BatchNorm


def test_merge_batchnorm_consistency(module_linear, module_batchnorm, data_linear):
    '''Test whether the output of the merged batchnorm is consistent with its original output.'''
    output_linear_before = module_linear(data_linear)
    output_batchnorm_before = module_batchnorm(output_linear_before)
    canonizer = SequentialMergeBatchNorm()

    try:
        canonizer.register((module_linear,), module_batchnorm)
        output_linear_canonizer = module_linear(data_linear)
        output_batchnorm_canonizer = module_batchnorm(output_linear_canonizer)
    finally:
        canonizer.remove()

    output_linear_after = module_linear(data_linear)
    output_batchnorm_after = module_batchnorm(output_linear_after)

    assert all(torch.allclose(left, right, atol=1e-5) for left, right in [
        (output_linear_before, output_linear_after),
        (output_batchnorm_before, output_batchnorm_after),
        (output_batchnorm_before, output_linear_canonizer),
        (output_linear_canonizer, output_batchnorm_canonizer),
    ])


@pytest.mark.parametrize('canonizer_fn', [
    SequentialMergeBatchNorm,
    partial(NamedMergeBatchNorm, [(['dense0'], 'bnorm0')]),
])
def test_merge_batchnorm_apply(canonizer_fn, module_linear, module_batchnorm, data_linear):
    '''Test whether SequentialMergeBatchNorm merges BatchNorm modules correctly and keeps the output unchanged.'''
    model = Sequential(OrderedDict([
        ('dense0', module_linear),
        ('bnorm0', module_batchnorm)
    ]))
    output_before = model(data_linear)

    handles = RemovableHandleList(
        module.register_forward_hook(assert_identity_hook(True, 'BatchNorm was not merged!'))
        for module in model.modules() if isinstance(module, BatchNorm)
    )

    canonizer = canonizer_fn()

    canonizer_handles = RemovableHandleList(canonizer.apply(model))
    try:
        output_canonizer = model(data_linear)
    finally:
        handles.remove()
        canonizer_handles.remove()

    handles = RemovableHandleList(
        module.register_forward_hook(assert_identity_hook(False, 'BatchNorm was not restored!'))
        for module in model.modules() if isinstance(module, BatchNorm)
    )

    try:
        output_after = model(data_linear)
    finally:
        handles.remove()

    assert torch.allclose(output_canonizer, output_before, rtol=1e-5), 'Canonizer changed output after register!'
    assert torch.allclose(output_before, output_after, rtol=1e-5), 'Canonizer changed output after remove!'


def test_attribute_canonizer(module_linear, data_linear):
    '''Test whether AttributeCanonizer overwrites and restores a linear module's forward correctly. '''
    model = Sequential(OrderedDict([
        ('dense0', module_linear),
    ]))
    output_before = model(data_linear)

    modules = [module_linear]
    module_type = type(module_linear)

    assert all(
        module.forward == module_type.forward.__get__(module) for module in modules
    ), 'Model has its forward already overwritten!'

    def attribute_map(name, module):
        if module is module_linear:
            return {'forward': lambda x: module_type.forward.__get__(module)(x) * 2}
        return None

    canonizer = AttributeCanonizer(attribute_map)

    handles = RemovableHandleList(canonizer.apply(model))
    try:
        assert not any(
            module.forward == module_type.forward.__get__(module) for module in modules
        ), 'Model forward was not overwritten!'
        output_canonizer = model(data_linear)
    finally:
        handles.remove()

    output_after = model(data_linear)

    assert all(
        module.forward == module_type.forward.__get__(module) for module in modules
    ), 'Model forward was not restored!'
    assert torch.allclose(output_canonizer, output_before * 2, rtol=1e-5), 'Canonizer output mismatch after register!'
    assert torch.allclose(output_before, output_after, rtol=1e-5), 'Canonizer changed output after remove!'


def test_composite_canonizer():
    '''Test whether CompositeCanonizer correctly combines two AttributeCanonizer canonizers.'''
    module_vanilla = torch.nn.Module()
    model = torch.nn.Sequential(module_vanilla)

    canonizer = CompositeCanonizer([
        AttributeCanonizer(lambda name, module: {'_test_x': 13}),
        AttributeCanonizer(lambda name, module: {'_test_y': 13}),
    ])

    handles = RemovableHandleList(canonizer.apply(model))
    try:
        assert hasattr(module_vanilla, '_test_x'), 'Model attribute _test_x was not overwritten!'
        assert hasattr(module_vanilla, '_test_y'), 'Model attribute _test_y was not overwritten!'
    finally:
        handles.remove()

    assert not hasattr(module_vanilla, '_test_x'), 'Model attribute _test_x was not removed!'
    assert not hasattr(module_vanilla, '_test_y'), 'Model attribute _test_y was not removed!'


def test_base_canonizer_cooperative_apply():
    '''Test whether Canonizer's apply method is cooperative.'''

    class DummyCanonizer(Canonizer):
        '''Class to test Canonizer's cooperative apply.'''
        def apply(self, root_module):
            '''Cooperative apply which appends a string 'dummy' to the result of the parent class.'''
            instances = super().apply(root_module)
            instances += ['dummy']
            return instances

        def register(self):
            '''No-op register for abstract method.'''

        def remove(self):
            '''No-op remove for abstract method.'''

    canonizer = DummyCanonizer()
    model = Sequential()
    instances = canonizer.apply(model)
    assert 'dummy' in instances, 'Unexpected canonizer instance list!'



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/rules.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Rules based on Hooks'''
import torch

from .core import Hook, BasicHook, Stabilizer, expand, ParamMod


def zero_bias(zero_params=None):
    '''Add `'bias'` to `zero_params`, where `zero_params` is a string or a list of strings.

    Parameters
    ----------
    zero_params: str or list of str, optional
        Name or names, to which ``'bias'`` should be added.

    Returns
    -------
    list of str
        Supplied ``zero_params``, with the string ``'bias'`` appended.
    '''
    if zero_params is None:
        return ['bias']
    if isinstance(zero_params, str):
        zero_params = [zero_params]
    if 'bias' in zero_params:
        return zero_params
    return list(zero_params) + ['bias']


class ClampMod(ParamMod):
    '''ParamMod to clamp module parameters.

    Parameters
    ----------
    min: float or None, optional
        Minimum float value for which the parameters should be clamped, or None if no clamping should be done.
    max: float or None, optional
        Maximum float value for which the parameters should be clamped, or None if no clamping should be done.
    kwargs: dict[str, object]
        Additional keyword arguments used for :py:class:`ParamMod`.
    '''
    def __init__(self, min=None, max=None, **kwargs):
        def modifier(param, name):
            return param.clamp(min=min, max=max)
        super().__init__(modifier, **kwargs)


class GammaMod(ParamMod):
    '''ParamMod to modify module parameters as in the Gamma rule. Adds the scaled, clamped parameters to the parameter
    itself.

    Parameters
    ----------
    gamma: float, optional
        Gamma scaling parameter, by which the clamped parameters are multiplied.
    min: float or None, optional
        Minimum float value for which the parameters should be clamped, or None if no clamping should be done.
    max: float or None, optional
        Maximum float value for which the parameters should be clamped, or None if no clamping should be done.
    kwargs: dict[str, object]
        Additional keyword arguments used for :py:class:`ParamMod`.
    '''
    def __init__(self, gamma=0.25, min=None, max=None, **kwargs):
        def modifier(param, name):
            return param + gamma * param.clamp(min=min, max=max)
        super().__init__(modifier, **kwargs)


class NoMod(ParamMod):
    '''ParamMod that does not modify the parameters. Allows other modification flags.

    Parameters
    ----------
    kwargs: dict[str, object]
        Additional keyword arguments used for :py:class:`ParamMod`.
    '''
    def __init__(self, **kwargs):
        super().__init__((lambda param, _: param), **kwargs)


class Epsilon(BasicHook):
    '''LRP Epsilon rule :cite:p:`bach2015pixel`.
    Setting ``(epsilon=0)`` produces the LRP-0 rule :cite:p:`bach2015pixel`.
    LRP Epsilon is most commonly used in middle layers, LRP-0 is most commonly used in upper layers
    :cite:p:`montavon2019layer`.
    Sometimes higher values of ``epsilon`` are used, therefore it is not always only a stabilizer value.

    Parameters
    ----------
    epsilon: callable or float, optional
        Stabilization parameter. If ``epsilon`` is a float, it will be added to the denominator with the same sign as
        each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected, of
        which the output corresponds to the stabilized denominator. Note that this is called ``stabilizer`` for all
        other rules.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    '''
    def __init__(self, epsilon=1e-6, zero_params=None):
        stabilizer_fn = Stabilizer.ensure(epsilon)
        super().__init__(
            input_modifiers=[lambda input: input],
            param_modifiers=[NoMod(zero_params=zero_params)],
            output_modifiers=[lambda output: output],
            gradient_mapper=(lambda out_grad, outputs: out_grad / stabilizer_fn(outputs[0])),
            reducer=(lambda inputs, gradients: inputs[0] * gradients[0]),
        )


class Gamma(BasicHook):
    '''Generalized LRP Gamma rule :cite:p:`montavon2019layer,andeol2021learning`.
    The gamma parameter scales the added positive/negative parts of the weights.
    The original Gamma rule :cite:p:`montavon2019layer` may only be used with positive inputs.
    The generalized version is equivalent to the original Gamma when there are only positive inputs, but may also be
    used for negative inputs.

    Parameters
    ----------
    gamma: float, optional
        Multiplier for added positive weights.
    stabilizer: callable or float, optional
        Stabilization parameter. If ``stabilizer`` is a float, it will be added to the denominator with the same sign
        as each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected,
        of which the output corresponds to the stabilized denominator.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    '''
    def __init__(self, gamma=0.25, stabilizer=1e-6, zero_params=None):
        mod_kwargs = {'zero_params': zero_params}
        mod_kwargs_nobias = {'zero_params': zero_bias(zero_params)}
        stabilizer_fn = Stabilizer.ensure(stabilizer)
        super().__init__(
            input_modifiers=[
                lambda input: input.clamp(min=0),
                lambda input: input.clamp(max=0),
                lambda input: input.clamp(min=0),
                lambda input: input.clamp(max=0),
                lambda input: input,
            ],
            param_modifiers=[
                GammaMod(gamma, min=0., **mod_kwargs),
                GammaMod(gamma, max=0., **mod_kwargs_nobias),
                GammaMod(gamma, max=0., **mod_kwargs),
                GammaMod(gamma, min=0., **mod_kwargs_nobias),
                NoMod(),
            ],
            output_modifiers=[lambda output: output] * 5,
            gradient_mapper=(
                lambda out_grad, outputs: [
                    output * out_grad / stabilizer_fn(denom)
                    for output, denom in (
                        [(outputs[4] > 0., sum(outputs[:2]))] * 2
                        + [(outputs[4] < 0., sum(outputs[2:4]))] * 2
                    )
                ] + [torch.zeros_like(out_grad)]
            ),
            reducer=(
                lambda inputs, gradients: sum(input * gradient for input, gradient in zip(inputs[:4], gradients[:4]))
            ),
        )


class ZPlus(BasicHook):
    '''LRP ZPlus rule :cite:p:`bach2015pixel,montavon2017explaining`.
    It is the same as using :py:class:`~zennit.rules.AlphaBeta` with ``(alpha=1, beta=0)``

    Parameters
    ----------
    stabilizer: callable or float, optional
        Stabilization parameter. If ``stabilizer`` is a float, it will be added to the denominator with the same sign
        as each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected,
        of which the output corresponds to the stabilized denominator.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.

    Notes
    -----
    Note that the original deep Taylor Decomposition (DTD) specification of the ZPlus Rule
    :cite:p:`montavon2017explaining` only considers positive inputs, as they are used in ReLU Networks.
    This implementation is effectively alpha=1, beta=0, where negative inputs are allowed.
    '''
    def __init__(self, stabilizer=1e-6, zero_params=None):
        stabilizer_fn = Stabilizer.ensure(stabilizer)
        super().__init__(
            input_modifiers=[
                lambda input: input.clamp(min=0),
                lambda input: input.clamp(max=0),
            ],
            param_modifiers=[
                ClampMod(min=0., zero_params=zero_params),
                ClampMod(max=0., zero_params=zero_bias(zero_params)),
            ],
            output_modifiers=[lambda output: output] * 2,
            gradient_mapper=(lambda out_grad, outputs: [out_grad / stabilizer_fn(sum(outputs))] * 2),
            reducer=(lambda inputs, gradients: inputs[0] * gradients[0] + inputs[1] * gradients[1]),
        )


class AlphaBeta(BasicHook):
    '''LRP AlphaBeta rule :cite:p:`bach2015pixel`.
    The AlphaBeta rule weights positive (alpha) and negative (beta) contributions.
    Most common parameters are ``(alpha=1, beta=0)`` and ``(alpha=2, beta=1)``.
    It is most commonly used for lower layers :cite:p:`montavon2019layer`.

    Parameters
    ----------
    alpha: float, optional
        Multiplier for the positive output term.
    beta: float, optional
        Multiplier for the negative output term.
    stabilizer: callable or float, optional
        Stabilization parameter. If ``stabilizer`` is a float, it will be added to the denominator with the same sign
        as each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected,
        of which the output corresponds to the stabilized denominator.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.

    '''
    def __init__(self, alpha=2., beta=1., stabilizer=1e-6, zero_params=None):
        if alpha < 0 or beta < 0:
            raise ValueError("Both alpha and beta parameters must be positive!")
        if (alpha - beta) != 1.:
            raise ValueError("The difference of parameters alpha - beta must equal 1!")
        mod_kwargs = {'zero_params': zero_params}
        mod_kwargs_nobias = {'zero_params': zero_bias(zero_params)}
        stabilizer_fn = Stabilizer.ensure(stabilizer)

        super().__init__(
            input_modifiers=[
                lambda input: input.clamp(min=0),
                lambda input: input.clamp(max=0),
                lambda input: input.clamp(min=0),
                lambda input: input.clamp(max=0),
            ],
            param_modifiers=[
                ClampMod(min=0., **mod_kwargs),
                ClampMod(max=0., **mod_kwargs_nobias),
                ClampMod(max=0., **mod_kwargs),
                ClampMod(min=0., **mod_kwargs_nobias),
            ],
            output_modifiers=[lambda output: output] * 4,
            gradient_mapper=(
                lambda out_grad, outputs: [
                    out_grad / stabilizer_fn(denom)
                    for denom in ([sum(outputs[:2])] * 2 + [sum(outputs[2:])] * 2)
                ]
            ),
            reducer=(
                lambda inputs, gradients: (
                    alpha * (inputs[0] * gradients[0] + inputs[1] * gradients[1])
                    - beta * (inputs[2] * gradients[2] + inputs[3] * gradients[3])
                )
            ),
        )


class ZBox(BasicHook):
    '''LRP ZBox rule :cite:p:`montavon2017explaining`.
    The ZBox rule is intended for "boxed" input pixel space.
    Generally, the lowest and highest *possible* values are used, i.e. ``(low=0., high=1.)`` for raw image data in
    the float data type.
    Neural network inputs are often normalized to match an isotropic gaussian distribution with mean 0 and variance 1,
    which means that the lowest and highest values also need to be adapted.
    For image data, this generally happens per channel, for which case ``low`` and ``high`` can be passed as tensors
    with shape ``(1, 3, 1, 1)``, which will be broadcasted as expected.

    Parameters
    ----------
    low: :py:class:`torch.Tensor` or float
        Lowest pixel values of input. Subject to broadcasting.
    high: :py:class:`torch.Tensor` or float
        Highest pixel values of input. Subject to broadcasting.
    stabilizer: callable or float, optional
        Stabilization parameter. If ``stabilizer`` is a float, it will be added to the denominator with the same sign
        as each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected,
        of which the output corresponds to the stabilized denominator.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.

    '''
    def __init__(self, low, high, stabilizer=1e-6, zero_params=None):
        def sub(positive, *negatives):
            return positive - sum(negatives)

        mod_kwargs = {'zero_params': zero_params}
        stabilizer_fn = Stabilizer.ensure(stabilizer)

        super().__init__(
            input_modifiers=[
                lambda input: input,
                lambda input: expand(low, input.shape, cut_batch_dim=True).to(input),
                lambda input: expand(high, input.shape, cut_batch_dim=True).to(input),
            ],
            param_modifiers=[
                NoMod(**mod_kwargs),
                ClampMod(min=0., **mod_kwargs),
                ClampMod(max=0., **mod_kwargs),
            ],
            output_modifiers=[lambda output: output] * 3,
            gradient_mapper=(lambda out_grad, outputs: (out_grad / stabilizer_fn(sub(*outputs)),) * 3),
            reducer=(lambda inputs, gradients: sub(*(input * gradient for input, gradient in zip(inputs, gradients)))),
        )


class Pass(Hook):
    '''Unmodified pass-through rule.
    If the rule of a layer shall not be any other, is elementwise and shall not be the gradient, the `Pass` rule simply
    passes upper layer relevance through to the lower layer.
    '''
    def backward(self, module, grad_input, grad_output):
        '''Pass through the upper gradient, skipping the one for this layer.'''
        return grad_output


class Norm(BasicHook):
    '''Normalize and weight by input contribution.
    This is essentially the same as the LRP :py:class:`~zennit.rules.Epsilon` rule :cite:p:`bach2015pixel` with a fixed
    epsilon only used as a stabilizer, and without the need of the attached layer to have parameters ``weight`` and
    ``bias``.
    '''
    def __init__(self, stabilizer=1e-6):
        stabilizer_fn = Stabilizer.ensure(stabilizer)
        super().__init__(
            input_modifiers=[lambda input: input],
            param_modifiers=[NoMod(param_keys=[])],
            output_modifiers=[lambda output: output],
            gradient_mapper=(lambda out_grad, outputs: out_grad / stabilizer_fn(outputs[0])),
            reducer=(lambda inputs, gradients: inputs[0] * gradients[0]),
        )


class WSquare(BasicHook):
    '''LRP WSquare rule :cite:p:`montavon2017explaining`.
    It is most commonly used in the first layer when the values are not bounded :cite:p:`montavon2019layer`.

    Parameters
    ----------
    stabilizer: callable or float, optional
        Stabilization parameter. If ``stabilizer`` is a float, it will be added to the denominator with the same sign
        as each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected,
        of which the output corresponds to the stabilized denominator.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    '''
    def __init__(self, stabilizer=1e-6, zero_params=None):
        stabilizer_fn = Stabilizer.ensure(stabilizer)
        super().__init__(
            input_modifiers=[torch.ones_like],
            param_modifiers=[
                ParamMod((lambda param, _: param ** 2), zero_params=zero_params),
            ],
            output_modifiers=[lambda output: output],
            gradient_mapper=(lambda out_grad, outputs: out_grad / stabilizer_fn(outputs[0])),
            reducer=(lambda inputs, gradients: gradients[0]),
        )


class Flat(BasicHook):
    '''LRP Flat rule :cite:p:`lapuschkin2019unmasking`.
    It is essentially the same as the LRP :py:class:`~zennit.rules.WSquare` rule, but with all parameters set to ones.

    Parameters
    ----------
    stabilizer: callable or float, optional
        Stabilization parameter. If ``stabilizer`` is a float, it will be added to the denominator with the same sign
        as each respective entry. If it is callable, a function ``(input: torch.Tensor) -> torch.Tensor`` is expected,
        of which the output corresponds to the stabilized denominator.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    '''
    def __init__(self, stabilizer=1e-6, zero_params=None):
        mod_kwargs = {'zero_params': zero_bias(zero_params), 'require_params': False}
        stabilizer_fn = Stabilizer.ensure(stabilizer)
        super().__init__(
            input_modifiers=[torch.ones_like],
            param_modifiers=[
                ParamMod((lambda param, name: torch.ones_like(param)), **mod_kwargs),
            ],
            output_modifiers=[lambda output: output],
            gradient_mapper=(lambda out_grad, outputs: out_grad / stabilizer_fn(outputs[0])),
            reducer=(lambda inputs, gradients: gradients[0]),
        )


class ReLUDeconvNet(Hook):
    '''DeconvNet ReLU rule :cite:p:`zeiler2014visualizing`.'''
    def backward(self, module, grad_input, grad_output):
        '''Modify ReLU gradient according to DeconvNet :cite:p:`zeiler2014visualizing`.'''
        return (grad_output[0].clamp(min=0),)


class ReLUGuidedBackprop(Hook):
    '''GuidedBackprop ReLU rule :cite:p:`springenberg2015striving`.'''
    def backward(self, module, grad_input, grad_output):
        '''Modify ReLU gradient according to GuidedBackprop :cite:p:`springenberg2015striving`.'''
        return (grad_input[0] * (grad_output[0] > 0.),)


class ReLUBetaSmooth(Hook):
    '''Modify ReLU gradient to smooth softplus gradient :cite:p:`dombrowski2019explanations`. Used to obtain meaningful
    surrogate gradients to compute higher order gradients with ReLUs. Equivalent to changing the gradient to be the
    (scaled) logistic function (sigmoid).

    Parameters
    ----------
    beta_smooth: float, optional
        The beta parameter for the softplus gradient (i.e. ``sigmoid(beta * input)``). Defaults to ``10``.
    '''
    def __init__(self, beta_smooth=10.):
        super().__init__()
        self.beta_smooth = beta_smooth

    def copy(self):
        '''Return a copy of this hook with the same beta parameter.'''
        return self.__class__(beta_smooth=self.beta_smooth)

    def forward(self, module, input, output):
        '''Remember the input for the backward pass.'''
        self.stored_tensors['input'] = input

    def backward(self, module, grad_input, grad_output):
        '''Modify ReLU gradient to the smooth softplus gradient :cite:p:`dombrowski2019explanations`.'''
        return (torch.sigmoid(self.beta_smooth * self.stored_tensors['input'][0]) * grad_output[0],)



'''Tests for composites using torchvision models.'''
from types import MethodType
from itertools import product

from zennit.core import BasicHook, collect_leaves


def ishookcopy(hook, hook_template):
    '''Check if ``hook`` is a copy of ``hook_template`` (due to copying-mechanics of BasicHook).'''
    if isinstance(hook_template, BasicHook):
        return all(
            getattr(hook, key) == getattr(hook_template, key)
            for key in (
                'input_modifiers',
                'param_modifiers',
                'output_modifiers',
                'gradient_mapper',
            )
        )
    return isinstance(hook, type(hook_template))


def check_hook_registered(module, hook_template):
    '''Check whether a ``hook_template`` has been registered to ``module``. '''
    return any(
        ishookcopy(hook_func.__self__, hook_template)
        for hook_func in getattr(module, '_forward_pre_hooks').values()
        if isinstance(hook_func, MethodType)
    )


def verify_no_hooks(model):
    '''Verify that ``model`` has no registered forward (-pre) hooks.'''
    return not any(
        any(getattr(module, key) for key in ('_forward_hooks', '_forward_pre_hooks'))
        for module in model.modules()
    )


def test_composite_layer_map_registered(layer_map_composite, model_vision):
    '''Tests whether the explicit LayerMapComposites register and unregister their rules correctly.'''
    errors = []
    with layer_map_composite.context(model_vision):
        for child in model_vision.modules():
            for dtype, hook_template in layer_map_composite.layer_map:
                if isinstance(child, dtype):
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first of {} but {} is not registered!',
                            (child, dtype, hook_template),
                        ))
                    break

    if not verify_no_hooks(model_vision):
        errors.append(('Model has hooks registered after composite was removed!', ()))

    assert not errors, 'Errors:\n  ' + '\n  '.join(f'[{n}] ' + msg.format(*arg) for n, (msg, arg) in enumerate(errors))


def test_composite_special_first_layer_map_registered(special_first_layer_map_composite, model_vision):
    '''Tests whether the explicit LayerMapComposites register and unregister their rules correctly.'''
    errors = []
    try:
        special_first_layer, special_first_template, special_first_dtype = next(
            (child, hook_template, dtype)
            for child, (dtype, hook_template) in product(
                collect_leaves(model_vision), special_first_layer_map_composite.first_map
            ) if isinstance(child, dtype)
        )
    except StopIteration:
        special_first_layer = None
        special_first_template = None

    with special_first_layer_map_composite.context(model_vision):
        if special_first_layer is not None and not check_hook_registered(special_first_layer, special_first_template):
            errors.append((
                'Special first layer {} is first of {} but {} is not registered!',
                (special_first_layer, special_first_dtype, special_first_template)
            ))

        children = (child for child in model_vision.modules() if child is not special_first_layer)
        for child in children:
            for dtype, hook_template in special_first_layer_map_composite.layer_map:
                if isinstance(child, dtype):
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first of {} but {} is not registered!',
                            (child, dtype, hook_template),
                        ))
                    break

    if not verify_no_hooks(model_vision):
        errors.append(('Model has hooks registered after composite was removed!', ()))

    assert not errors, 'Errors:\n  ' + '\n  '.join(f'[{n}] ' + msg.format(*arg) for n, (msg, arg) in enumerate(errors))


def test_composite_name_map_registered(name_map_composite, model_vision):
    '''Tests whether the constructed NameMapComposites register and unregister their rules correctly.'''
    errors = []
    with name_map_composite.context(model_vision):
        for name, child in model_vision.named_modules():
            for names, hook_template in name_map_composite.name_map:
                if name in names:
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first in name map for {}, but is not registered!',
                            (name, hook_template),
                        ))
                    break

    if not verify_no_hooks(model_vision):
        errors.append(('Model has hooks registered after composite was removed!', ()))

    assert not errors, 'Errors:\n  ' + '\n  '.join(f'[{n}] ' + msg.format(*arg) for n, (msg, arg) in enumerate(errors))


def test_composite_mixed_registered(mixed_composite, model_vision):
    '''Tests whether the constructed MixedComposites register and unregister their rules correctly.'''
    errors = []

    name_map_composite, special_first_layer_map_composite = mixed_composite.composites

    try:
        special_first_layer, special_first_template, special_first_dtype = next(
            (child, hook_template, dtype)
            for child, (dtype, hook_template) in product(
                collect_leaves(model_vision), special_first_layer_map_composite.first_map
            ) if isinstance(child, dtype)
        )
    except StopIteration:
        special_first_layer = None
        special_first_template = None

    with mixed_composite.context(model_vision):
        has_matched_first_layer = False
        for name, child in model_vision.named_modules():
            has_matched_name_map = False
            for names, hook_template in name_map_composite.name_map:
                if name in names:
                    has_matched_name_map = True
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first in name map for {}, but is not registered!',
                            (name, hook_template),
                        ))
                    break

            if has_matched_name_map:
                continue

            if not has_matched_first_layer and child == special_first_layer:
                has_matched_first_layer = True
                if not check_hook_registered(child, special_first_template):
                    errors.append((
                        'Special first layer {} is first of {} but {} is not registered!',
                        (special_first_layer, special_first_dtype, special_first_template)
                    ))
                continue

            for dtype, hook_template in special_first_layer_map_composite.layer_map:
                if isinstance(child, dtype):
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first of {} but {} is not registered!',
                            (child, dtype, hook_template),
                        ))
                    break

    if not verify_no_hooks(model_vision):
        errors.append(('Model has hooks registered after composite was removed!', ()))

    assert not errors, 'Errors:\n  ' + '\n  '.join(f'[{n}] ' + msg.format(*arg) for n, (msg, arg) in enumerate(errors))


def test_composite_name_layer_map_registered(name_layer_map_composite, model_vision):
    '''Tests whether the constructed NameLayerMapComposites register and unregister their rules correctly.'''
    errors = []

    name_map_composite, layer_map_composite = name_layer_map_composite.composites

    with name_layer_map_composite.context(model_vision):
        for name, child in model_vision.named_modules():
            for names, hook_template in name_map_composite.name_map:
                has_matched_name_map = False
                if name in names:
                    has_matched_name_map = True
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first in name map for {}, but is not registered!',
                            (name, hook_template),
                        ))
                    break

            if has_matched_name_map:
                continue

            for dtype, hook_template in layer_map_composite.layer_map:
                if isinstance(child, dtype):
                    if not check_hook_registered(child, hook_template):
                        errors.append((
                            '{} is first of {} but {} is not registered!',
                            (child, dtype, hook_template),
                        ))
                    break

    if not verify_no_hooks(model_vision):
        errors.append(('Model has hooks registered after composite was removed!', ()))

    assert not errors, 'Errors:\n  ' + '\n  '.join(f'[{n}] ' + msg.format(*arg) for n, (msg, arg) in enumerate(errors))



#!/usr/bin/env python3
import re
from setuptools import setup, find_packages
from subprocess import run, CalledProcessError


def get_long_description(project_path):
    '''Fetch the README contents and replace relative links with absolute ones
    pointing to github for correct behaviour on PyPI.
    '''
    try:
        revision = run(
            ['git', 'describe', '--tags'],
            capture_output=True,
            check=True,
            text=True
        ).stdout[:-1]
    except CalledProcessError:
        try:
            with open('PKG-INFO', 'r') as fd:
                body = fd.read().partition('\n\n')[2]
            if body:
                return body
        except FileNotFoundError:
            revision = 'master'

    with open('README.md', 'r', encoding='utf-8') as fd:
        long_description = fd.read()

    link_root = {
        '': f'https://github.com/{project_path}/blob',
        '!': f'https://raw.githubusercontent.com/{project_path}',
    }

    def replace(mobj):
        return f'{mobj[1]}[{mobj[2]}]({link_root[mobj[1]]}/{revision}/{mobj[3]})'

    link_rexp = re.compile(r'(!?)\[([^\]]*)\]\((?!https?://|/)([^\)]+)\)')
    return link_rexp.sub(replace, long_description)


setup(
    name='zennit',
    use_scm_version=True,
    author='chrstphr',
    author_email='zennit@j0d.de',
    description='Attribution of Neural Networks using PyTorch',
    long_description=get_long_description('chr5tphr/zennit'),
    long_description_content_type='text/markdown',
    url='https://github.com/chr5tphr/zennit',
    packages=find_packages(where='src', include=['zennit*']),
    package_dir={'': 'src'},
    install_requires=[
        'click',
        'numpy',
        'Pillow',
        'torch>=1.7.0',
        'torchvision',
    ],
    setup_requires=[
        'setuptools_scm',
    ],
    extras_require={
        'docs': [
            'sphinx-copybutton>=0.4.0',
            'sphinx-rtd-theme>=1.0.0',
            'sphinxcontrib.datatemplates>=0.9.0',
            'sphinxcontrib.bibtex>=2.4.1',
            'nbsphinx>=0.8.8',
            'ipykernel>=6.13.0',
        ],
        'tests': [
            'pytest',
            'pytest-cov',
        ]
    },
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
)



'''Tests for ColorMap and CMSL.'''
from typing import NamedTuple
import pytest
import numpy as np

from zennit.cmap import ColorMap, LazyColorMapCache


class CMapExample(NamedTuple):
    '''Named tuple for example color maps used in tests.'''
    source: str
    nodes: list


CMAPS = [
    ('000,fff', [
        (0x00, (0x00, 0x00, 0x00)),
        (0xff, (0xff, 0xff, 0xff)),
    ]),
    ('fff,f00', [
        (0x00, (0xff, 0xff, 0xff)),
        (0xff, (0xff, 0x00, 0x00)),
    ]),
    ('fff,00f', [
        (0x00, (0xff, 0xff, 0xff)),
        (0xff, (0x00, 0x00, 0xff)),
    ]),
    ('000,f00,ff0,fff', [
        (0x00, (0x00, 0x00, 0x00)),
        (0x55, (0xff, 0x00, 0x00)),
        (0xaa, (0xff, 0xff, 0x00)),
        (0xff, (0xff, 0xff, 0xff)),
    ]),
    ('000,00f,0ff', [
        (0x00, (0x00, 0x00, 0x00)),
        (0x7f, (0x00, 0x00, 0xff)),
        (0xff, (0x00, 0xff, 0xff)),
    ]),
    ('0ff,00f,80:000,f00,ff0,fff', [
        (0x00, (0x00, 0xff, 0xff)),
        (0x40, (0x00, 0x00, 0xff)),
        (0x80, (0x00, 0x00, 0x00)),
        (0xaa, (0xff, 0x00, 0x00)),
        (0xd4, (0xff, 0xff, 0x00)),
        (0xff, (0xff, 0xff, 0xff)),
    ]),
    ('00f,80:fff,f00', [
        (0x00, (0x00, 0x00, 0xff)),
        (0x80, (0xff, 0xff, 0xff)),
        (0xff, (0xff, 0x00, 0x00)),
    ]),
    ('0055a4,80:fff,ef4135', [
        (0x00, (0x00, 0x55, 0xa4)),
        (0x80, (0xff, 0xff, 0xff)),
        (0xff, (0xef, 0x41, 0x35)),
    ]),
    ('0000d0,80:d0d0d0,d00000', [
        (0x00, (0x00, 0x00, 0xd0)),
        (0x80, (0xd0, 0xd0, 0xd0)),
        (0xff, (0xd0, 0x00, 0x00)),
    ]),
    ('00d0d0,80:d0d0d0,d000d0', [
        (0x00, (0x00, 0xd0, 0xd0)),
        (0x80, (0xd0, 0xd0, 0xd0)),
        (0xff, (0xd0, 0x00, 0xd0)),
    ]),
    ('00d000,80:d0d0d0,d000d0', [
        (0x00, (0x00, 0xd0, 0x00)),
        (0x80, (0xd0, 0xd0, 0xd0)),
        (0xff, (0xd0, 0x00, 0xd0)),
    ]),
    ('7:000, 9:ffffff', [
        (0x00, (0x00, 0x00, 0x00)),
        (0x77, (0x00, 0x00, 0x00)),
        (0x99, (0xff, 0xff, 0xff)),
        (0xff, (0xff, 0xff, 0xff)),
    ]),
]


def interpolate(x, nodes):
    '''Interpolate from example color map nodes.'''
    xp_addr = np.array([node[0] for node in nodes], dtype=np.float64)
    fp_rgb = np.array([node[1] for node in nodes], dtype=np.float64).T
    return np.stack([np.interp(x, xp_addr, fp) for fp in fp_rgb], axis=-1).round(12).clip(0., 255.).astype(np.uint8)


@pytest.fixture(scope='session', params=CMAPS)
def cmap_example(request):
    '''Example color map fixture.'''
    return CMapExample(*request.param)


@pytest.mark.parametrize('source_code', [
    'this', 'fff', ',,,', '111:111:111', 'fffff,fffff', 'f,f', 'fffffffff', 'ff:', 'ff:fff,00:000'
])
def test_color_map_wrong_syntax(source_code):
    '''Test whether different kinds of syntax errors cause a RuntimeError.'''
    with pytest.raises(RuntimeError):
        ColorMap(source_code)


def test_color_map_nodes_call(cmap_example):
    '''Test if the color map nodes have the specified color when calling a ColorMap instance.'''
    cmap = ColorMap(cmap_example.source)
    input_addr = np.array([node[0] for node in cmap_example.nodes], dtype=np.float64)[None]
    expected_rgb = np.array([node[1] for node in cmap_example.nodes], dtype=np.uint8)[None]
    cmap_rgb = (cmap(input_addr / 255.) * 255.).round(12).clip(0., 255.).astype(np.uint8)
    assert np.allclose(expected_rgb, cmap_rgb)


def test_color_map_nodes_palette(cmap_example):
    '''Test if the color map nodes have the specified color when using ColorMap.palette.'''
    cmap = ColorMap(cmap_example.source)
    input_addr = [node[0] for node in cmap_example.nodes]
    expected_rgb = np.array([node[1] for node in cmap_example.nodes], dtype=np.uint8)[None]
    palette = cmap.palette(level=1.)
    cmap_rgb = palette[input_addr]
    assert np.allclose(expected_rgb, cmap_rgb)


def test_color_map_full_call(cmap_example):
    '''Test if the color map nodes have correctly interpolated colors when calling a ColorMap instance.'''
    cmap = ColorMap(cmap_example.source)
    input_addr = np.arange(256, dtype=np.uint8)
    expected_rgb = interpolate(input_addr, cmap_example.nodes)
    cmap_rgb = (cmap(input_addr / 255.) * 255.).round(12).clip(0., 255.).astype(np.uint8)
    assert np.allclose(expected_rgb, cmap_rgb)


def test_color_map_full_palette(cmap_example):
    '''Test if the color map nodes have correctly interpolated colors when using ColorMap.palette.'''
    input_addr = np.arange(256, dtype=np.uint8)
    expected_palette = interpolate(input_addr, cmap_example.nodes)
    cmap = ColorMap(cmap_example.source)
    cmap_palette = cmap.palette(level=1.0)
    assert np.allclose(expected_palette, cmap_palette)


def test_color_map_reassign_source_palette(cmap_example):
    '''Test if calling a ColorMap instance for which the source was changed produces correctly interpolated colors.'''
    cmap = ColorMap('fff,fff')
    cmap.source = cmap_example.source

    input_addr = np.arange(256, dtype=np.uint8)
    expected_palette = interpolate(input_addr, cmap_example.nodes)
    cmap_palette = cmap.palette(level=1.0)
    assert np.allclose(expected_palette, cmap_palette)


def test_color_map_source_property(cmap_example):
    '''Test if the source property of a color map is equal to the specified source code.'''
    cmap = ColorMap(cmap_example.source)
    assert cmap.source == cmap_example.source, 'Mismatching source!'


@pytest.fixture(scope='function')
def lazy_cmap_cache():
    '''Single fixture for a LazyColorMapCache'''
    return LazyColorMapCache({
        'gray': '000,fff',
        'red': '100,f00',
    })


class TestLazyColorMapCache:
    '''Tests for LazyColorMapCache.'''
    @staticmethod
    def test_missing(lazy_cmap_cache):
        '''Test whether accessing an unknown key causes a KeyError.'''
        with pytest.raises(KeyError):
            _ = lazy_cmap_cache['no such cmap']

    @staticmethod
    def test_get_item_uncompiled(lazy_cmap_cache):
        '''Test whether accessing an uncompiled entry compiles and returns the correct ColorMap.'''
        cmap = lazy_cmap_cache['red']
        assert isinstance(cmap, ColorMap)
        assert cmap.source == '100,f00'

    @staticmethod
    def test_get_item_cached(lazy_cmap_cache):
        '''Test whether accessing a previously compiled and cached entry returns the same ColorMap.'''
        cmaps = [
            lazy_cmap_cache['red'],
            lazy_cmap_cache['red'],
        ]
        assert cmaps[0] is cmaps[1]

    @staticmethod
    def test_set_item_existing(lazy_cmap_cache):
        '''Test whether setting an already existing, uncompiled entry and accessing it returns the correct ColorMap.'''
        lazy_cmap_cache['red'] = 'fff,f00'
        assert lazy_cmap_cache['red'].source == 'fff,f00'

    @staticmethod
    def test_set_item_new(lazy_cmap_cache):
        '''Test whether setting a new entry and accessing it returns the correct ColorMap.'''
        lazy_cmap_cache['blue'] = 'fff,00f'
        assert lazy_cmap_cache['blue'].source == 'fff,00f'

    @staticmethod
    def test_set_item_compiled(lazy_cmap_cache):
        '''Test whether setting an already existing, compiled entry and accessing it returns the same, modified
        ColorMap instance.
        '''
        original_cmap = lazy_cmap_cache['red']
        lazy_cmap_cache['red'] = 'fff,f00'
        assert lazy_cmap_cache['red'].source == 'fff,f00'
        assert original_cmap is lazy_cmap_cache['red']

    @staticmethod
    def test_del_item_uncompiled(lazy_cmap_cache):
        '''Test whether deleting an uncompiled entry correctly removes the entry.'''
        del lazy_cmap_cache['red']
        assert 'red' not in lazy_cmap_cache

    @staticmethod
    def test_del_item_compiled(lazy_cmap_cache):
        '''Test whether deleting a compiled entry correctly removes the entry.'''
        _ = lazy_cmap_cache['red']
        del lazy_cmap_cache['red']
        assert 'red' not in lazy_cmap_cache

    @staticmethod
    def test_iter(lazy_cmap_cache):
        '''Test whether iterating a LazyColorMapCache returns its keys.'''
        assert (list(lazy_cmap_cache) == ['gray', 'red'])

    @staticmethod
    def test_len(lazy_cmap_cache):
        '''Test whether calling len on a LazyColorMapCache returns the correct length.'''
        assert len(lazy_cmap_cache) == 2



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/canonizers.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Functions to produce a canonical form of models fit for LRP'''
from abc import ABCMeta, abstractmethod

import torch

from .core import collect_leaves
from .types import Linear, BatchNorm, ConvolutionTranspose


class Canonizer(metaclass=ABCMeta):
    '''Canonizer Base class.
    Canonizers modify modules temporarily such that certain attribution rules can properly be applied.
    '''
    @abstractmethod
    def apply(self, root_module):
        '''Apply this canonizer recursively on all applicable modules.

        Parameters
        ----------
        root_module: :py:obj:`torch.nn.Module`
            Root module to which to apply the canonizers.

        Returns
        -------
        list
            A list of all applied instances of this class.
        '''
        return []

    @abstractmethod
    def register(self):
        '''Apply the changes of this canonizer.'''

    @abstractmethod
    def remove(self):
        '''Revert the changes introduces by this canonizer.'''

    def copy(self):
        '''Return a copy of this instance.'''
        return self.__class__()


class MergeBatchNorm(Canonizer):
    '''Abstract Canonizer to merge the parameters of batch norms into linear modules.'''
    linear_type = (
        Linear,
    )
    batch_norm_type = (
        BatchNorm,
    )

    def __init__(self):
        super().__init__()
        self.linears = None
        self.batch_norm = None

        self.linear_params = None
        self.batch_norm_params = None

    def register(self, linears, batch_norm):
        '''Store the parameters of the linear modules and the batch norm module and apply the merge.

        Parameters
        ----------
        linear: list of :py:obj:`torch.nn.Module`
            List of linear layer with mandatory attributes `weight` and `bias`.
        batch_norm: :py:obj:`torch.nn.Module`
            Batch Normalization module with mandatory attributes
            `running_mean`, `running_var`, `weight`, `bias` and `eps`
        '''
        self.linears = linears
        self.batch_norm = batch_norm

        self.linear_params = [(linear.weight.data, getattr(linear.bias, 'data', None)) for linear in linears]

        self.batch_norm_params = {
            key: getattr(self.batch_norm, key).data for key in ('weight', 'bias', 'running_mean', 'running_var')
        }

        self.merge_batch_norm(self.linears, self.batch_norm)

    def remove(self):
        '''Undo the merge by reverting the parameters of both the linear and the batch norm modules to the state before
        the merge.
        '''
        for linear, (weight, bias) in zip(self.linears, self.linear_params):
            linear.weight.data = weight
            if bias is None:
                linear.bias = None
            else:
                linear.bias.data = bias

        for key, value in self.batch_norm_params.items():
            getattr(self.batch_norm, key).data = value

    @staticmethod
    def merge_batch_norm(modules, batch_norm):
        '''Update parameters of a linear layer to additionally include a Batch Normalization operation and update the
        batch normalization layer to instead compute the identity.

        Parameters
        ----------
        modules: list of :py:obj:`torch.nn.Module`
            Linear layers with mandatory attributes `weight` and `bias`.
        batch_norm: :py:obj:`torch.nn.Module`
            Batch Normalization module with mandatory attributes `running_mean`, `running_var`, `weight`, `bias` and
            `eps`
        '''
        denominator = (batch_norm.running_var + batch_norm.eps) ** .5
        scale = batch_norm.weight / denominator

        for module in modules:
            original_weight = module.weight.data
            if module.bias is None:
                module.bias = torch.nn.Parameter(
                    torch.zeros(1, device=original_weight.device, dtype=original_weight.dtype)
                )
            original_bias = module.bias.data

            if isinstance(module, ConvolutionTranspose):
                index = (None, slice(None), *((None,) * (original_weight.ndim - 2)))
            else:
                index = (slice(None), *((None,) * (original_weight.ndim - 1)))

            # merge batch_norm into linear layer
            module.weight.data = original_weight * scale[index]
            module.bias.data = (original_bias - batch_norm.running_mean) * scale + batch_norm.bias

        # change batch_norm parameters to produce identity
        batch_norm.running_mean.data = torch.zeros_like(batch_norm.running_mean.data)
        batch_norm.running_var.data = torch.ones_like(batch_norm.running_var.data)
        batch_norm.bias.data = torch.zeros_like(batch_norm.bias.data)
        batch_norm.weight.data = torch.ones_like(batch_norm.weight.data)


class SequentialMergeBatchNorm(MergeBatchNorm):
    '''Canonizer to merge the parameters of all batch norms that appear sequentially right after a linear module.

    Note
    ----
    SequentialMergeBatchNorm traverses the tree of children of the provided module depth-first and in-order.
    This means that child-modules must be assigned to their parent module in the order they are visited in the forward
    pass to correctly identify adjacent modules.
    This also means that activation functions must be assigned in their module-form as a child to their parent-module
    to properly detect when there is an activation function between linear and batch-norm modules.

    '''
    def apply(self, root_module):
        '''Finds a batch norm following right after a linear layer, and creates a copy of this instance to merge
        them by fusing the batch norm parameters into the linear layer and reducing the batch norm to the identity.

        Parameters
        ----------
        root_module: :py:obj:`torch.nn.Module`
            A module of which the leaves will be searched and if a batch norm is found right after a linear layer, will
            be merged.

        Returns
        -------
        instances: list
            A list of instances of this class which modified the appropriate leaves.
        '''
        instances = []
        last_leaf = None
        for leaf in collect_leaves(root_module):
            if isinstance(last_leaf, self.linear_type) and isinstance(leaf, self.batch_norm_type):
                instance = self.copy()
                instance.register((last_leaf,), leaf)
                instances.append(instance)
            last_leaf = leaf

        return instances


class NamedMergeBatchNorm(MergeBatchNorm):
    '''Canonizer to merge the parameters of all batch norms into linear modules, specified by their respective names.

    Parameters
    ----------
    name_map: list[tuple[string], string]
        List of which linear layer names belong to which batch norm name.
    '''
    def __init__(self, name_map):
        super().__init__()
        self.name_map = name_map

    def apply(self, root_module):
        '''Create appropriate merges given by the name map.

        Parameters
        ----------
        root_module: :py:obj:`torch.nn.Module`
            Root module for which underlying modules will be merged.

        Returns
        -------
        instances: list
            A list of merge instances.
        '''
        instances = []
        lookup = dict(root_module.named_modules())

        for linear_names, batch_norm_name in self.name_map:
            instance = self.copy()
            instance.register([lookup[name] for name in linear_names], lookup[batch_norm_name])
            instances.append(instance)

        return instances

    def copy(self):
        return self.__class__(self.name_map)


class AttributeCanonizer(Canonizer):
    '''Canonizer to set an attribute of module instances.
    Note that the use of this Canonizer removes previously set attributes after removal.

    Parameters
    ----------
    attribute_map: Function
        A function that returns either None, if not applicable, or a dict with keys describing which attributes to
        overload for a module. The function signature is (name: string, module: type) -> None or
        dict.
    '''
    def __init__(self, attribute_map):
        self.attribute_map = attribute_map
        self.attribute_keys = None
        self.module = None

    def apply(self, root_module):
        '''Overload the attributes for all applicable modules.

        Parameters
        ----------
        root_module: :py:obj:`torch.nn.Module`
            Root module for which underlying modules will have their attributes overloaded.

        Returns
        -------
        instances : list of :py:obj:`Canonizer`
            The applied canonizer instances, which may be removed by calling `.remove`.
        '''
        instances = []
        for name, module in root_module.named_modules():
            attributes = self.attribute_map(name, module)
            if attributes is not None:
                instance = self.copy()
                instance.register(module, attributes)
                instances.append(instance)
        return instances

    def register(self, module, attributes):
        '''Overload the module's attributes.

        Parameters
        ---------
        module : :py:obj:`torch.nn.Module`
            The module of which the attributes will be overloaded.
        attributes : dict
            The attributes which to overload for the module.
        '''
        self.attribute_keys = list(attributes)
        self.module = module
        for key, value in attributes.items():
            setattr(module, key, value)

    def remove(self):
        '''Remove the overloaded attributes. Note that functions are descriptors, and therefore not direct attributes
        of instance, which is why deleting instance attributes with the same name reverts them to the original
        function.
        '''
        for key in self.attribute_keys:
            delattr(self.module, key)

    def copy(self):
        '''Copy this Canonizer.

        Returns
        -------
        :py:obj:`Canonizer`
            A copy of this Canonizer.
        '''
        return AttributeCanonizer(self.attribute_map)


class CompositeCanonizer(Canonizer):
    '''A Composite of Canonizers, which applies all supplied canonizers.

    Parameters
    ----------
    canonizers : list of :py:obj:`Canonizer`
        Canonizers of which to build a Composite of.
    '''
    def __init__(self, canonizers):
        self.canonizers = canonizers

    def apply(self, root_module):
        '''Apply call canonizers.

        Parameters
        ----------
        root_module: :py:obj:`torch.nn.Module`
            Root module for which underlying modules will have canonizers applied.

        Returns
        -------
        instances : list of :py:obj:`Canonizer`
            The applied canonizer instances, which may be removed by calling `.remove`.
        '''
        instances = []
        for canonizer in self.canonizers:
            instances += canonizer.apply(root_module)
        return instances

    def register(self):
        '''Register this Canonizer. Nothing to do for a CompositeCanonizer.'''

    def remove(self):
        '''Remove this Canonizer. Nothing to do for a CompositeCanonizer.'''



'''Tests for Attributors.'''
from functools import partial
from itertools import product

import pytest
import torch

from zennit.attribution import Gradient, IntegratedGradients, SmoothGrad, Occlusion, occlude_independent


class IdentityLogger(torch.nn.Module):
    '''Helper-Module to log input tensors.'''
    def __init__(self):
        super().__init__()
        self.tensors = []

    def forward(self, input):
        '''Clone input, append to self.tensors and return the cloned tensor.'''
        self.tensors.append(input.clone())
        return self.tensors[-1]


def test_gradient_attributor_inactive(
    data_linear, model_simple, model_simple_output, any_composite, grad_outputs_func
):
    '''Test whether composite context and attributor match for Gradient.'''

    with Gradient(model=model_simple, composite=any_composite, attr_output=grad_outputs_func) as attributor:
        # verify that all hooks are active
        assert all(hook.active for hook in attributor.composite.hook_refs)
        with attributor.inactive():
            # verify that all hooks are inactive
            assert all(not hook.active for hook in attributor.composite.hook_refs)


def test_gradient_attributor_composite(
    data_linear, model_simple, model_simple_output, any_composite, grad_outputs_func
):
    '''Test whether composite context and attributor match for Gradient.'''
    with any_composite.context(model_simple) as module:
        data = data_linear.detach().requires_grad_()
        output_context = module(data)
        grad_outputs = grad_outputs_func(output_context)
        grad_context, = torch.autograd.grad(output_context, data, grad_outputs)

    with Gradient(model=model_simple, composite=any_composite, attr_output=grad_outputs_func) as attributor:
        output_attributor, grad_attributor = attributor(data_linear)

    assert torch.allclose(output_context, output_attributor)
    assert torch.allclose(grad_context, grad_attributor)
    assert torch.allclose(model_simple_output, output_attributor)


@pytest.mark.parametrize('use_const,use_call,use_init', product(*[[True, False]] * 3))
def test_gradient_attributor_output_fn(data_simple, grad_outputs_func, use_const, use_call, use_init):
    '''Test whether attributors' attr_output supports functions, constants and None in any of supplied or not supplied
    for each the attributor initialization and the call.
    '''
    model = IdentityLogger()

    attr_output = grad_outputs_func(data_simple) if use_const else grad_outputs_func
    init_attr_output = attr_output if use_init else None
    call_attr_output = attr_output if use_call else None

    with Gradient(model=model, attr_output=init_attr_output) as attributor:
        _, grad = attributor(data_simple, attr_output=call_attr_output)

    if (use_call or use_init):
        expected_grad = grad_outputs_func(data_simple)
    else:
        # the identity is the default attr_output
        expected_grad = data_simple

    assert torch.allclose(expected_grad, grad), 'Attributor output function gradient mismatch!'


def test_gradient_attributor_grad(data_simple):
    '''Test whether the gradient of Gradient matches.'''
    model = torch.nn.Softplus(beta=1.)
    data = data_simple.view_as(data_simple).requires_grad_()
    target = torch.sigmoid(data) * (1 - torch.sigmoid(data))

    with Gradient(model=model, create_graph=True) as attributor:
        _, grad = attributor(data, torch.ones_like)
    gradgrad, = torch.autograd.grad(grad.sum(), data)

    assert torch.allclose(gradgrad, target), 'Gradient Attributor second order gradient mismatch!'


def test_gradient_attributor_output_fn_precedence(data_simple):
    '''Test whether the gradient attributor attr_output at call is preferred when it is supplied at both initialization
    and call.
    '''
    model = IdentityLogger()

    init_attr_output = torch.ones_like
    call_attr_output = torch.zeros_like

    with Gradient(model=model, attr_output=init_attr_output) as attributor:
        _, grad = attributor(data_simple, attr_output=call_attr_output)

    expected_grad = call_attr_output(data_simple)
    assert torch.allclose(expected_grad, grad), 'Attributor output function precedence mismatch!'


def test_smooth_grad_single(data_linear, model_simple, model_simple_output, model_simple_grad):
    '''Test whether SmoothGrad with a single iteration is equal to the gradient.'''
    with SmoothGrad(model=model_simple, noise_level=0.1, n_iter=1) as attributor:
        output, grad = attributor(data_linear)

    assert torch.allclose(model_simple_grad, grad)
    assert torch.allclose(model_simple_output, output)


def test_smooth_grad_single_grad(data_simple):
    '''Test whether the gradient of SmoothGrad matches.'''
    model = torch.nn.Softplus(beta=1.)
    data = data_simple.view_as(data_simple).requires_grad_()
    target = torch.sigmoid(data) * (1 - torch.sigmoid(data))

    with SmoothGrad(model=model, noise_level=0.1, n_iter=1, create_graph=True) as attributor:
        _, grad = attributor(data, torch.ones_like)
    gradgrad, = torch.autograd.grad(grad.sum(), data)

    assert torch.allclose(gradgrad, target), 'SmoothGrad Attributor second order gradient mismatch!'


@pytest.mark.parametrize('noise_level', [0.0, 0.1, 0.3, 0.5])
def test_smooth_grad_distribution(data_simple, noise_level):
    '''Test whether the SmoothGrad sampled distribution matches.'''
    model = IdentityLogger()

    dims = tuple(range(1, data_simple.ndim))
    noise_var = (noise_level * (data_simple.amax(dims) - data_simple.amin(dims))) ** 2
    n_iter = 100

    with SmoothGrad(model=model, noise_level=noise_level, n_iter=n_iter, attr_output=torch.ones_like) as attributor:
        _, grad = attributor(data_simple)

    assert len(model.tensors) == n_iter, 'SmoothGrad iterations did not match n_iter!'

    sample_mean = sum(model.tensors) / len(model.tensors)
    sample_var = ((sum((tensor - sample_mean) ** 2 for tensor in model.tensors) / len(model.tensors))).mean(dims)

    if noise_level > 0.:
        std_ratio = (sample_var / noise_var) ** .5
        assert (std_ratio < 1.5).all().item(), 'SmoothGrad sample variance is too high!'
        assert (std_ratio > 0.667).all().item(), 'SmoothGrad sample variance is too low!'
    else:
        assert (sample_var < 1e-9).all().item(), 'SmoothGrad sample variance is too high!'
    assert torch.allclose(grad, torch.ones_like(data_simple)), 'SmoothGrad of identity is wrong!'


@pytest.mark.parametrize('baseline_fn', [None, torch.zeros_like, torch.ones_like])
def test_integrated_gradients_single(data_linear, model_simple, model_simple_output, model_simple_grad, baseline_fn):
    '''Test whether IntegratedGradients with a single iteration is equal to the expected output given multiple
    baselines.
    '''
    with IntegratedGradients(model=model_simple, n_iter=1, baseline_fn=baseline_fn) as attributor:
        output, grad = attributor(data_linear)

    if baseline_fn is None:
        baseline_fn = torch.zeros_like
    expected_grad = model_simple_grad * (data_linear - baseline_fn(data_linear))

    assert torch.allclose(expected_grad, grad), 'Gradient mismatch for IntegratedGradients!'
    assert torch.allclose(model_simple_output, output), 'Output mismatch for IntegratedGradients!'


def test_integrated_gradients_single_grad(data_simple):
    '''Test whether the gradient of IntegratedGradients matches.'''
    model = torch.nn.Softplus(beta=1.)
    data = data_simple.view_as(data_simple).requires_grad_()
    # this is d/dx (x * d/dx softplus(x)), i.e. the gradient of input times gradient of softplus
    target = torch.sigmoid(data) * (1 - torch.sigmoid(data)) * data + torch.sigmoid(data)

    with IntegratedGradients(model=model, n_iter=1, baseline_fn=torch.zeros_like, create_graph=True) as attributor:
        _, grad = attributor(data, torch.ones_like)
    gradgrad, = torch.autograd.grad(grad.sum(), data)

    assert torch.allclose(gradgrad, target), 'IntegratedGradients Attributor second order gradient mismatch!'


def test_integrated_gradients_path(data_simple):
    '''Test whether IntegratedGradients with a single iteration and a zero-baseline is equal to the input times the
    gradient.
    '''
    model = IdentityLogger()

    dims = tuple(range(1, data_simple.ndim))
    n_iter = 100
    with IntegratedGradients(model=model, n_iter=n_iter, attr_output=torch.ones_like) as attributor:
        _, grad = attributor(data_simple)

    assert len(model.tensors) == n_iter, 'IntegratedGradients iterations did not match n_iter!'

    data_simple_norm = data_simple / (data_simple ** 2).sum(dim=dims, keepdim=True) ** .5
    assert all(
        torch.allclose(step / (step ** 2).sum(dim=dims, keepdim=True) ** .5, data_simple_norm)
        for step in model.tensors
    ), 'IntegratedGradients segments do not lie on path!'
    assert torch.allclose(data_simple, grad), 'IntegratedGradients of identity is wrong!'


@pytest.mark.parametrize('window,stride', zip([1, 2, 4, (1,), (2,), (4,)], [1, 2, 4, (1,), (2,), (4,)]))
def test_occlusion_disjunct(data_simple, window, stride):
    '''Function to test whether the inputs used for disjunct occlusion windows are correct.'''
    model = IdentityLogger()

    # delete everything except the window
    occlusion_fn = partial(occlude_independent, fill_fn=torch.zeros_like, invert=False)

    with Occlusion(model=model, window=window, stride=stride, occlusion_fn=occlusion_fn) as attributor:
        attributor(data_simple)

    # omit final pass for full output
    reconstruct = sum(model.tensors[:-1])
    assert torch.allclose(data_simple, reconstruct), 'Disjunct occlusion does not sum to original input!'


@pytest.mark.parametrize(
    'fill_fn,invert', [
        (None, False),
        (torch.zeros_like, False),
        (torch.zeros_like, True),
        (torch.ones_like, True),
    ]
)
def test_occlusion_single(data_linear, model_simple, model_simple_output, grad_outputs_func, fill_fn, invert):
    '''Function to test whether the inputs used for a full occlusion window are correct.'''
    window, stride = [data_linear.shape] * 2
    if fill_fn is None:
        # setting when no occlusion_fn is supplied
        occlusion_fn = None
        fill_fn = torch.zeros_like
    else:
        occlusion_fn = partial(occlude_independent, fill_fn=fill_fn, invert=invert)

    identity_logger = IdentityLogger()
    model = torch.nn.Sequential(identity_logger, model_simple)

    with Occlusion(
        model=model,
        window=window,
        stride=stride,
        attr_output=grad_outputs_func,
        occlusion_fn=occlusion_fn,
    ) as attributor:
        output, score = attributor(data_linear)

    expected_occluded = fill_fn(data_linear) if invert else data_linear
    expected_output = model_simple(expected_occluded)
    expected_score = grad_outputs_func(expected_output).sum(
        tuple(range(1, expected_output.ndim))
    )[(slice(None),) + (None,) * (data_linear.ndim - 1)].expand_as(data_linear)

    assert len(identity_logger.tensors) == 2, 'Incorrect number of forward passes for Occlusion!'
    assert torch.allclose(identity_logger.tensors[0], expected_occluded), 'Occluded input mismatch!'
    assert torch.allclose(model_simple_output, output), 'Output mismatch for Occlusion!'
    assert torch.allclose(expected_score, score), 'Scores are incorrect for Occlusion!'


@pytest.mark.parametrize('argument,container', product(
    ['window', 'stride'],
    ['monkey', {3}, ('you', 'are', 'breathtaking'), range(3), [3]]
))
def test_occlusion_stride_window_typecheck(argument, container):
    '''Test whether Occlusion raises a TypeError on incorrect types for window and stride.'''
    with pytest.raises(TypeError):
        Occlusion(model=None, **{argument: container})



'''Helper functions for various tests.'''
from itertools import product

import numpy as np
import torch

from zennit.types import BatchNorm


def prodict(**kwargs):
    '''Create a dictionary with values which are the cartesian product of the input keyword arguments.'''
    return [dict(zip(kwargs, val)) for val in product(*kwargs.values())]


def one_hot_max(output):
    '''Get the one-hot encoded max.'''
    return torch.sparse_coo_tensor(
        [*zip(np.unravel_index(output.argmax(), output.shape))], [1.], output.shape, dtype=output.dtype
    ).to_dense()


def assert_identity_hook(equal=True, message=''):
    '''Create an assertion hook which checks whether the module does or does not modify its input.'''
    def assert_identity(module, input, output):
        '''Assert whether the module does or does not modify its input.'''
        assert equal == torch.allclose(input[0], output, rtol=1e-5), message
    return assert_identity


def randomize_bnorm(model):
    '''Randomize all BatchNorm module parameters of a model.'''
    for module in model.modules():
        if isinstance(module, BatchNorm):
            module.weight.data.uniform_(0.1, 2.0)
            module.running_var.data.uniform_(0.1, 2.0)
            module.bias.data.normal_()
            module.running_mean.data.normal_()
            # smaller eps to reduce error
            module.eps = 1e-30
    return model


def nograd(model):
    '''Unset grad requirement for all model parameters.'''
    for param in model.parameters():
        param.requires_grad = False
    return model



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/composites.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Composites, registered in a global composite dict.'''
import torch

from .core import Composite
from .layer import Sum
from .rules import Gamma, Epsilon, ZBox, ZPlus, AlphaBeta, Flat, Pass, Norm
from .rules import ReLUDeconvNet, ReLUGuidedBackprop, ReLUBetaSmooth
from .types import Convolution, Linear, AvgPool, Activation, BatchNorm


class LayerMapComposite(Composite):
    '''A Composite for which hooks are specified by a mapping from module types to hooks.

    Parameters
    ----------
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, layer_map, canonizers=None):
        self.layer_map = layer_map
        super().__init__(module_map=self.mapping, canonizers=canonizers)

    # pylint: disable=unused-argument
    def mapping(self, ctx, name, module):
        '''Get the appropriate hook given a mapping from module types to hooks.

        Parameters
        ----------
        ctx: dict
            A context dictionary to keep track of previously registered hooks.
        name: str
            Name of the module.
        module: obj:`torch.nn.Module`
            Instance of the module to find a hook for.

        Returns
        -------
        obj:`Hook` or None
            The hook found with the module type in the given layer map,
            or None if no applicable hook was found.
        '''
        return next((hook for types, hook in self.layer_map if isinstance(module, types)), None)


class SpecialFirstLayerMapComposite(LayerMapComposite):
    '''A Composite for which hooks are specified by a mapping from module types to hooks.

    Parameters
    ----------
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook.
    first_map: `list[tuple[tuple[torch.nn.Module, ...], Hook]]`
        Applicable mapping for the first layer, same format as `layer_map`.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, layer_map, first_map, canonizers=None):
        self.first_map = first_map
        super().__init__(layer_map=layer_map, canonizers=canonizers)

    def mapping(self, ctx, name, module):
        '''Get the appropriate hook given a mapping from module types to hooks and a special mapping for the first
        occurrence in another mapping.

        Parameters
        ----------
        ctx: dict
            A context dictionary to keep track of previously registered hooks.
        name: str
            Name of the module.
        module: obj:`torch.nn.Module`
            Instance of the module to find a hook for.

        Returns
        -------
        obj:`Hook` or None
            The hook found with the module type in the given layer map,
            in the first layer map if this was the first layer,
            or None if no applicable hook was found.
        '''
        if not ctx.get('first_layer_visited', False):
            for types, hook in self.first_map:
                if isinstance(module, types):
                    ctx['first_layer_visited'] = True
                    return hook

        return super().mapping(ctx, name, module)


class NameMapComposite(Composite):
    '''A Composite for which hooks are specified by a mapping from module names to hooks.

    Parameters
    ----------
    name_map: `list[tuple[tuple[str, ...], Hook]]`
        A mapping as a list of tuples, with a tuple of applicable module names and a Hook.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, name_map, canonizers=None):
        self.name_map = name_map
        super().__init__(module_map=self.mapping, canonizers=canonizers)

    # pylint: disable=unused-argument
    def mapping(self, ctx, name, module):
        '''Get the appropriate hook given a mapping from module names to hooks.

        Parameters
        ----------
        ctx: dict
            A context dictionary to keep track of previously registered hooks.
        name: str
            Name of the module.
        module: obj:`torch.nn.Module`
            Instance of the module to find a hook for.

        Returns
        -------
        obj:`Hook` or None
            The hook found with the module name in the given name map, or None if no applicable
            hook was found.
        '''
        return next((hook for names, hook in self.name_map if name in names), None)


class MixedComposite(Composite):
    '''A Composite for which hooks are specified by a list of composites.

    Each composite defines a mapping from layer property to a specific Hook.
    The list order of composites defines their matching order.

    Parameters
    ----------
    composites: `list[Composite]`
        A list of Composites. The list order of composites defines their matching order.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, composites, canonizers=None):
        if canonizers is None:
            canonizers = []
        self.composites = composites
        super().__init__(
            module_map=self.mapping,
            canonizers=sum([composite.canonizers for composite in composites], canonizers)
        )

    def mapping(self, ctx, name, module):
        '''Get the appropriate hook given a list of composites.

        Parameters
        ----------
        ctx: dict
            A context dictionary to keep track of previously registered hooks.
        name: str
            Name of the module.
        module: obj:`torch.nn.Module`
            Instance of the module to find a hook for.

        Returns
        -------
        obj:`Hook` or None
            The hook found by the first match in the composite list,
            or None if no applicable hook was found.
        '''
        # create a context for each of the sub-composites if ctx is empty
        if not ctx:
            ctx.update({composite: {} for composite in self.composites})

        # create list of hooks by evaluating module maps of all given composites
        hooks = [composite.module_map(ctx[composite], name, module) for composite in self.composites]

        # return first hook that is not None, if there isn't any, return None
        return next((hook for hook in hooks if hook is not None), None)


class NameLayerMapComposite(MixedComposite):
    '''A Composite for which hooks are specified by both a mapping from
    module names and module types to hooks.

    This implicitly creates instances of NameMapComposite and LayerMapComposite.
    The layer-name mapping will be matched before the layer-type mapping.

    Parameters
    ----------
    name_map: `list[tuple[tuple[str, ...], Hook]]`
        A mapping as a list of tuples, with a tuple of applicable module names and a Hook.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]], optional
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, name_map=None, layer_map=None, canonizers=None):
        super().__init__(composites=[
            NameMapComposite(name_map=name_map),
            LayerMapComposite(layer_map=layer_map),
        ], canonizers=canonizers)


COMPOSITES = {}


def register_composite(name):
    '''Register a composite in the global COMPOSITES dict under `name`.'''
    def wrapped(composite):
        '''Wrapped function to be called on the composite to register it to the global COMPOSITES dict.'''
        COMPOSITES[name] = composite
        return composite
    return wrapped


def layer_map_base(stabilizer=1e-6):
    '''Return a basic layer map (list of 2-tuples) shared by all built-in LayerMapComposites.

    Parameters
    ----------
    stabilizer: callable or float, optional
        Stabilization parameter for rules other than ``Epsilon``. If ``stabilizer`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator.

    Returns
    -------
    list[tuple[tuple[torch.nn.Module, ...], Hook]]
        Basic ayer map shared by all built-in LayerMapComposites.
    '''
    return [
        (Activation, Pass()),
        (Sum, Norm(stabilizer=stabilizer)),
        (AvgPool, Norm(stabilizer=stabilizer)),
        (BatchNorm, Pass()),
    ]


@register_composite('epsilon_gamma_box')
class EpsilonGammaBox(SpecialFirstLayerMapComposite):
    '''An explicit composite using the ZBox rule for the first convolutional layer, gamma rule for all following
    convolutional layers, and the epsilon rule for all fully connected layers.

    Parameters
    ----------
    low: obj:`torch.Tensor`
        A tensor with the same size as the input, describing the lowest possible pixel values.
    high: obj:`torch.Tensor`
        A tensor with the same size as the input, describing the highest possible pixel values.
    epsilon: callable or float, optional
        Stabilization parameter for the ``Epsilon`` rule. If ``epsilon`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator. Note that this is
        called ``stabilizer`` for all other rules.
    gamma: float, optional
        Gamma parameter for the gamma rule.
    stabilizer: callable or float, optional
        Stabilization parameter for rules other than ``Epsilon``. If ``stabilizer`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    first_map: `list[tuple[tuple[torch.nn.Module, ...], Hook]]`
        Applicable mapping for the first layer, same format as `layer_map`. This will be prepended to the ``first_map``
        defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(
        self,
        low,
        high,
        epsilon=1e-6,
        gamma=0.25,
        stabilizer=1e-6,
        layer_map=None,
        first_map=None,
        zero_params=None,
        canonizers=None
    ):
        if layer_map is None:
            layer_map = []
        if first_map is None:
            first_map = []

        rule_kwargs = {'zero_params': zero_params}
        layer_map = layer_map + layer_map_base(stabilizer) + [
            (Convolution, Gamma(gamma=gamma, stabilizer=stabilizer, **rule_kwargs)),
            (torch.nn.Linear, Epsilon(epsilon=epsilon, **rule_kwargs)),
        ]
        first_map = first_map + [
            (Convolution, ZBox(low=low, high=high, stabilizer=stabilizer, **rule_kwargs))
        ]
        super().__init__(layer_map=layer_map, first_map=first_map, canonizers=canonizers)


@register_composite('epsilon_plus')
class EpsilonPlus(LayerMapComposite):
    '''An explicit composite using the zplus rule for all convolutional layers and the epsilon rule for all fully
    connected layers.

    Parameters
    ----------
    epsilon: callable or float, optional
        Stabilization parameter for the ``Epsilon`` rule. If ``epsilon`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator. Note that this is
        called ``stabilizer`` for all other rules.
    stabilizer: callable or float, optional
        Stabilization parameter for rules other than ``Epsilon``. If ``stabilizer`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, epsilon=1e-6, stabilizer=1e-6, layer_map=None, zero_params=None, canonizers=None):
        if layer_map is None:
            layer_map = []

        rule_kwargs = {'zero_params': zero_params}
        layer_map = layer_map + layer_map_base(stabilizer) + [
            (Convolution, ZPlus(stabilizer=stabilizer, **rule_kwargs)),
            (torch.nn.Linear, Epsilon(epsilon=epsilon, **rule_kwargs)),
        ]
        super().__init__(layer_map=layer_map, canonizers=canonizers)


@register_composite('epsilon_alpha2_beta1')
class EpsilonAlpha2Beta1(LayerMapComposite):
    '''An explicit composite using the alpha2-beta1 rule for all convolutional layers and the epsilon rule for all
    fully connected layers.

    Parameters
    ----------
    epsilon: callable or float, optional
        Stabilization parameter for the ``Epsilon`` rule. If ``epsilon`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator. Note that this is
        called ``stabilizer`` for all other rules.
    stabilizer: callable or float, optional
        Stabilization parameter for rules other than ``Epsilon``. If ``stabilizer`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, epsilon=1e-6, stabilizer=1e-6, layer_map=None, zero_params=None, canonizers=None):
        if layer_map is None:
            layer_map = []

        rule_kwargs = {'zero_params': zero_params}
        layer_map = layer_map + layer_map_base(stabilizer) + [
            (Convolution, AlphaBeta(alpha=2, beta=1, stabilizer=stabilizer, **rule_kwargs)),
            (torch.nn.Linear, Epsilon(epsilon=epsilon, **rule_kwargs)),
        ]
        super().__init__(layer_map=layer_map, canonizers=canonizers)


@register_composite('epsilon_plus_flat')
class EpsilonPlusFlat(SpecialFirstLayerMapComposite):
    '''An explicit composite using the flat rule for any linear first layer, the zplus rule for all other convolutional
    layers and the epsilon rule for all other fully connected layers.

    Parameters
    ----------
    epsilon: callable or float, optional
        Stabilization parameter for the ``Epsilon`` rule. If ``epsilon`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator. Note that this is
        called ``stabilizer`` for all other rules.
    stabilizer: callable or float, optional
        Stabilization parameter for rules other than ``Epsilon``. If ``stabilizer`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    first_map: `list[tuple[tuple[torch.nn.Module, ...], Hook]]`
        Applicable mapping for the first layer, same format as `layer_map`. This will be prepended to the ``first_map``
        defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(
        self, epsilon=1e-6, stabilizer=1e-6, layer_map=None, first_map=None, zero_params=None, canonizers=None
    ):
        if layer_map is None:
            layer_map = []
        if first_map is None:
            first_map = []

        rule_kwargs = {'zero_params': zero_params}
        layer_map = layer_map + layer_map_base(stabilizer) + [
            (Convolution, ZPlus(stabilizer=stabilizer, **rule_kwargs)),
            (torch.nn.Linear, Epsilon(epsilon=epsilon, **rule_kwargs)),
        ]
        first_map = first_map + [
            (Linear, Flat(stabilizer=stabilizer, **rule_kwargs))
        ]
        super().__init__(layer_map=layer_map, first_map=first_map, canonizers=canonizers)


@register_composite('epsilon_alpha2_beta1_flat')
class EpsilonAlpha2Beta1Flat(SpecialFirstLayerMapComposite):
    '''An explicit composite using the flat rule for any linear first layer, the alpha2-beta1 rule for all other
    convolutional layers and the epsilon rule for all other fully connected layers.

    Parameters
    ----------
    epsilon: callable or float, optional
        Stabilization parameter for the ``Epsilon`` rule. If ``epsilon`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator. Note that this is
        called ``stabilizer`` for all other rules.
    stabilizer: callable or float, optional
        Stabilization parameter for rules other than ``Epsilon``. If ``stabilizer`` is a float, it will be added to the
        denominator with the same sign as each respective entry. If it is callable, a function ``(input: torch.Tensor)
        -> torch.Tensor`` is expected, of which the output corresponds to the stabilized denominator.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    first_map: `list[tuple[tuple[torch.nn.Module, ...], Hook]]`
        Applicable mapping for the first layer, same format as `layer_map`. This will be prepended to the ``first_map``
        defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(
        self, epsilon=1e-6, stabilizer=1e-6, layer_map=None, first_map=None, zero_params=None, canonizers=None
    ):
        if layer_map is None:
            layer_map = []
        if first_map is None:
            first_map = []

        rule_kwargs = {'zero_params': zero_params}
        layer_map = layer_map + layer_map_base(stabilizer) + [
            (Convolution, AlphaBeta(alpha=2, beta=1, stabilizer=stabilizer, **rule_kwargs)),
            (torch.nn.Linear, Epsilon(epsilon=epsilon, **rule_kwargs)),
        ]
        first_map = first_map + [
            (Linear, Flat(stabilizer=stabilizer, **rule_kwargs))
        ]
        super().__init__(layer_map=layer_map, first_map=first_map, canonizers=canonizers)


@register_composite('deconvnet')
class DeconvNet(LayerMapComposite):
    '''An explicit composite modifying the gradients of all ReLUs according to DeconvNet
    :cite:p:`zeiler2014visualizing`.

    Parameters
    ----------
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, layer_map=None, canonizers=None):
        if layer_map is None:
            layer_map = []

        layer_map = layer_map + [
            (torch.nn.ReLU, ReLUDeconvNet()),
        ]
        super().__init__(layer_map, canonizers=canonizers)


@register_composite('guided_backprop')
class GuidedBackprop(LayerMapComposite):
    '''An explicit composite modifying the gradients of all ReLUs according to GuidedBackprop
    :cite:p:`springenberg2015striving`.

    Parameters
    ----------
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, layer_map=None, canonizers=None):
        if layer_map is None:
            layer_map = []

        layer_map = layer_map + [
            (torch.nn.ReLU, ReLUGuidedBackprop()),
        ]
        super().__init__(layer_map=layer_map, canonizers=canonizers)


@register_composite('excitation_backprop')
class ExcitationBackprop(LayerMapComposite):
    '''An explicit composite implementing the ExcitationBackprop :cite:p:`zhang2016top`.

    Parameters
    ----------
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, stabilizer=1e-6, layer_map=None, zero_params=None, canonizers=None):
        if layer_map is None:
            layer_map = []

        layer_map = layer_map + [
            (Sum, Norm(stabilizer=stabilizer)),
            (AvgPool, Norm(stabilizer=stabilizer)),
            (Linear, ZPlus(stabilizer=stabilizer, zero_params=zero_params)),
        ]
        super().__init__(layer_map=layer_map, canonizers=canonizers)


@register_composite('beta_smooth')
class BetaSmooth(LayerMapComposite):
    '''Explicit composite to modify ReLU gradients to smooth softplus gradients :cite:p:`dombrowski2019explanations`.
    Used to obtain meaningful surrogate gradients to compute higher order gradients with ReLUs. Equivalent to changing
    the gradient to be the (scaled) logistic function (sigmoid).

    Parameters
    ----------
    beta_smooth: float, optional
        The beta parameter for the softplus gradient (i.e. ``sigmoid(beta * input)``). Defaults to ``10``.
    layer_map: list[tuple[tuple[torch.nn.Module, ...], Hook]]
        A mapping as a list of tuples, with a tuple of applicable module types and a Hook. This will be prepended to
        the ``layer_map`` defined by the composite.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, beta_smooth=10., layer_map=None, zero_params=None, canonizers=None):
        if layer_map is None:
            layer_map = []

        layer_map = layer_map + [
            (torch.nn.ReLU, ReLUBetaSmooth(beta_smooth=beta_smooth)),
        ]
        super().__init__(layer_map=layer_map, canonizers=canonizers)



'''Tests for various rules. Rules are re-implemented in a slower, less complicated way, which closely follows the
definition in the original works, which makes them easier to compare and thus less likely to be wrong.
'''
from functools import wraps, partial
from copy import deepcopy

import pytest
import torch
from zennit.rules import Epsilon, ZPlus, AlphaBeta, Gamma, ZBox, Norm, WSquare, Flat
from zennit.rules import Pass, ReLUDeconvNet, ReLUGuidedBackprop, ReLUBetaSmooth
from zennit.rules import zero_bias as name_zero_bias


def stabilize(input, epsilon=1e-6):
    '''Replicates zennit.core.stabilize for testing.'''
    return input + ((input == 0.).to(input) + input.sign()) * epsilon


def as_matrix(module_linear, input, output):
    '''Get flat weight and bias using the jacobian.'''
    jac = torch.autograd.functional.jacobian(module_linear, input[None])
    weight = jac.reshape((output.numel(), input.numel()))
    bias = output.flatten() - weight @ input.flatten()
    return weight, bias


RULES_LINEAR = []
RULES_SIMPLE = []


def replicates(target_list, replicated_func, **kwargs):
    '''Decorator to indicate a replication of a function for testing.'''
    def wrapper(func):
        '''Append to ``RULES_LINEAR`` as partial, given ``kwargs``.'''
        target_list.append(
            pytest.param(
                (partial(replicated_func, **kwargs), partial(func, **kwargs)),
                id=replicated_func.__name__
            )
        )
        return func
    return wrapper


def flat_module_params(func):
    '''Decorator to to copy module and overwrite module params completely with ones (for rule_flat).'''
    @wraps(func)
    def wrapped(module_linear, *args, **kwargs):
        '''Make a deep copy of module_linear, fill all parameters inline with ones, and call func with the copy.'''
        module_copy = deepcopy(module_linear)
        for param in module_copy.parameters():
            param.requires_grad_(False).fill_(1.0)
        return func(module_copy, *args, **kwargs)
    return wrapped


def matrix_form(func):
    '''Decorator to wrap function such that weights and bias supplied in matrix-form and input and output are flattened
    appropriately.'''
    @wraps(func)
    def wrapped(module_linear, input, output, **kwargs):
        '''Get flat weight matrix and bias using the jacobian, flatten input and output, and pass arguments to func.'''
        weight, bias = as_matrix(module_linear, input[0], output[0])
        return func(
            weight,
            bias,
            input.flatten(start_dim=1),
            output.flatten(start_dim=1),
            **kwargs
        ).reshape(input.shape)
    return wrapped


def with_grad(func):
    '''Decorator to wrap function such that the gradient is computed and passed to the function instead of module.'''
    @wraps(func)
    def wrapped(module, input, output, **kwargs):
        '''Get gradient and pass along input, output and keyword arguments to func.'''
        gradient, = torch.autograd.grad(module(input), input, output)
        return func(
            gradient,
            input,
            output,
            **kwargs
        )
    return wrapped


def zero_bias(zero_params, bias):
    '''Return a tensor with zeros like ``bias`` if zero_params is equal to or contains the string ``'bias'``, otherwise
    return the unmodified tensor ``bias``.'''
    if zero_params is None:
        zero_params = []
    if bias is not None and (zero_params == 'bias' or 'bias' in zero_params):
        return torch.zeros_like(bias)
    return bias


@replicates(RULES_LINEAR, Epsilon, epsilon=1e-6)
@replicates(RULES_LINEAR, Epsilon, epsilon=1e-6, zero_params='bias')
@replicates(RULES_LINEAR, Epsilon, epsilon=1.0)
@replicates(RULES_LINEAR, Epsilon, epsilon=1.0, zero_params='bias')
@replicates(RULES_LINEAR, Norm)
@matrix_form
def rule_epsilon(weight, bias, input, relevance, epsilon=1e-6, zero_params=None):
    '''Replicates the Epsilon rule.'''
    bias = zero_bias(zero_params, bias)
    return input * ((relevance / stabilize(input @ weight.t() + bias, epsilon)) @ weight)


@replicates(RULES_LINEAR, ZPlus)
@replicates(RULES_LINEAR, ZPlus, zero_params='bias')
@matrix_form
def rule_zplus(weight, bias, input, relevance, zero_params=None):
    '''Replicates the ZPlus rule.'''
    bias = zero_bias(zero_params, bias)
    wplus = weight.clamp(min=0)
    wminus = weight.clamp(max=0)
    xplus = input.clamp(min=0)
    xminus = input.clamp(max=0)
    zval = xplus @ wplus.t() + xminus @ wminus.t() + bias.clamp(min=0)
    rfac = relevance / stabilize(zval)
    return xplus * (rfac @ wplus) + xminus * (rfac @ wminus)


@replicates(RULES_LINEAR, Gamma, gamma=0.25)
@replicates(RULES_LINEAR, Gamma, gamma=0.25, zero_params='bias')
@replicates(RULES_LINEAR, Gamma, gamma=0.5)
@replicates(RULES_LINEAR, Gamma, gamma=0.5, zero_params='bias')
@matrix_form
def rule_gamma(weight, bias, input, relevance, gamma, zero_params=None):
    '''Replicates the Gamma rule.'''
    output = input @ weight.t() + bias
    bias = zero_bias(zero_params, bias)
    pinput = input.clamp(min=0)
    ninput = input.clamp(max=0)
    pwgamma = weight + weight.clamp(min=0) * gamma
    nwgamma = weight + weight.clamp(max=0) * gamma
    pbgamma = bias + bias.clamp(min=0) * gamma
    nbgamma = bias + bias.clamp(max=0) * gamma

    pgrad_out = (relevance / stabilize(pinput @ pwgamma.t() + ninput @ nwgamma.t() + pbgamma)) * (output > 0.)
    positive = pinput * (pgrad_out @ pwgamma) + ninput * (pgrad_out @ nwgamma)

    ngrad_out = (relevance / stabilize(pinput @ nwgamma.t() + ninput @ pwgamma.t() + nbgamma)) * (output < 0.)
    negative = pinput * (ngrad_out @ nwgamma) + ninput * (ngrad_out @ pwgamma)

    return positive + negative


@replicates(RULES_LINEAR, AlphaBeta, alpha=2.0, beta=1.0)
@replicates(RULES_LINEAR, AlphaBeta, alpha=1.0, beta=0.0, zero_params='bias')
@replicates(RULES_LINEAR, AlphaBeta, alpha=2.0, beta=1.0)
@replicates(RULES_LINEAR, AlphaBeta, alpha=1.0, beta=0.0, zero_params='bias')
@matrix_form
def rule_alpha_beta(weight, bias, input, relevance, alpha, beta, zero_params=None):
    '''Replicates the AlphaBeta rule.'''
    bias = zero_bias(zero_params, bias)
    wplus = weight.clamp(min=0)
    wminus = weight.clamp(max=0)
    xplus = input.clamp(min=0)
    xminus = input.clamp(max=0)
    zalpha = xplus @ wplus.t() + xminus @ wminus.t() + bias.clamp(min=0)
    zbeta = xplus @ wminus.t() + xminus @ wplus.t() + bias.clamp(max=0)
    ralpha = relevance / stabilize(zalpha)
    rbeta = relevance / stabilize(zbeta)
    result_alpha = xplus * (ralpha @ wplus) + xminus * (ralpha @ wminus)
    result_beta = xplus * (rbeta @ wminus) + xminus * (rbeta @ wplus)
    return alpha * result_alpha - beta * result_beta


@replicates(RULES_LINEAR, ZBox, low=-3.0, high=3.0)
@replicates(RULES_LINEAR, ZBox, low=-3.0, high=3.0, zero_params='bias')
@matrix_form
def rule_zbox(weight, bias, input, relevance, low, high, zero_params=None):
    '''Replicates the ZBox rule.'''
    wplus = weight.clamp(min=0)
    wminus = weight.clamp(max=0)
    low = torch.tensor(low).expand_as(input).to(input)
    high = torch.tensor(high).expand_as(input).to(input)
    zval = input @ weight.t() - low @ wplus.t() - high @ wminus.t()
    rfac = relevance / stabilize(zval)
    return input * (rfac @ weight) - low * (rfac @ wplus) - high * (rfac @ wminus)


@replicates(RULES_LINEAR, WSquare)
@replicates(RULES_LINEAR, WSquare, zero_params='bias')
@matrix_form
def rule_wsquare(weight, bias, input, relevance, zero_params=None):
    '''Replicates the WSquare rule.'''
    bias = zero_bias(zero_params, bias)
    wsquare = weight ** 2
    zval = torch.ones_like(input) @ wsquare.t() + bias ** 2
    rfac = relevance / stabilize(zval)
    return rfac @ wsquare


@replicates(RULES_LINEAR, Flat)
@flat_module_params
@matrix_form
def rule_flat(wflat, bias, input, relevance):
    '''Replicates the Flat rule.'''
    zval = torch.ones_like(input) @ wflat.t()
    rfac = relevance / stabilize(zval)
    return rfac @ wflat


@replicates(RULES_SIMPLE, Pass)
def rule_pass(module, input, relevance):
    '''Replicates the Pass rule.'''
    return relevance


@replicates(RULES_SIMPLE, ReLUDeconvNet)
def rule_relu_deconvnet(module, input, relevance):
    '''Replicates the ReLUDeconvNet rule.'''
    return relevance.clamp(min=0)


@replicates(RULES_SIMPLE, ReLUGuidedBackprop)
@with_grad
def rule_relu_guidedbackprop(gradient, input, relevance):
    '''Replicates the ReLUGuidedBackprop rule.'''
    return gradient * (relevance > 0.)


@replicates(RULES_SIMPLE, ReLUBetaSmooth, beta_smooth=10.)
@replicates(RULES_SIMPLE, ReLUBetaSmooth, beta_smooth=1.)
def rule_relu_beta_smooth(module, input, relevance, beta_smooth):
    '''Replicates the ReLUBetaSmooth rule.'''
    return relevance * torch.sigmoid(beta_smooth * input)


@pytest.fixture(scope='session', params=RULES_LINEAR)
def rule_pair_linear(request):
    '''Fixture to supply ``RULES_LINEAR``.'''
    return request.param


@pytest.fixture(scope='session', params=RULES_SIMPLE)
def rule_pair_simple(request):
    '''Fixture to supply ``RULES_SIMPLE``.'''
    return request.param


def compare_rule_pair(module, data, rule_pair):
    '''Compare rules with their replicated versions.'''
    rule_hook, rule_replicated = rule_pair

    input = data.clone().requires_grad_()
    handle = rule_hook().register(module)
    try:
        output = module(input)
        relevance_hook, = torch.autograd.grad(output, input, grad_outputs=output)
    finally:
        handle.remove()

    relevance_replicated = rule_replicated(module, input, output)

    assert torch.allclose(relevance_hook, relevance_replicated, atol=1e-5)


def test_linear_rule(module_linear, data_linear, rule_pair_linear):
    '''Test whether replicated and original implementations of rules for linear layers agree.'''
    compare_rule_pair(module_linear, data_linear, rule_pair_linear)


def test_simple_rule(module_simple, data_simple, rule_pair_simple):
    '''Test whether replicated and original implementations of rules for simple layers agree.'''
    compare_rule_pair(module_simple, data_simple, rule_pair_simple)


def test_alpha_beta_invalid_values():
    '''Test whether AlphaBeta raises ValueErrors for negative alpha/beta or when alpha - beta is not equal to 1.'''
    with pytest.raises(ValueError):
        AlphaBeta(alpha=-1.)
    with pytest.raises(ValueError):
        AlphaBeta(beta=-1.)
    with pytest.raises(ValueError):
        AlphaBeta(alpha=1., beta=1.)


@pytest.mark.parametrize('params', [None, 'weight', ['weight'], 'bias', ['bias'], ['weight', 'bias']])
def test_zero_bias(params):
    '''Test whether zero_bias correctly appends 'bias' to the zero_params list/str used for ParamMod.'''
    result = name_zero_bias(params)
    assert isinstance(result, list)
    assert 'bias' in result
